﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;

namespace MyECULib
{
    public class MyECUData
    {
        public enum types { Invalid, CompensationInfo, AnalogInputs, TimingInfo, DiagnosticInfo, VersionInfo };
        public types type = types.Invalid;
        public enum voltages { Volt5, Volt12 }; 
        public MyECUDataCompensationInfo ci;
        public MyECUDataAnalogInputs ai;
        public MyECUDataTimingInfo ti;
        public MyECUDataDiagnosticInfo di;
        public MyECUDataVersionInfo vi;

        public static double voltage2Lambda(double volt)
        {
            return volt * MyECUSource.LC1lambdaMultiplier + MyECUSource.LC1zero;
        }

        public static double lambda2AFR(double lambda)
        {
            return lambda * MyECUSource.AFRmultiplier;
        }

        public static uint hexToUInt(string str)
        {
            // Adjust for send format byte order
            if (str.Length != 4)
            {
                throw new Exception("Wrong length");
            }
            return uint.Parse(str.Substring(2, 2) + str.Substring(0, 2), System.Globalization.NumberStyles.HexNumber);
        }

        public static Int16 hexToInt(string str)
        {
            // Adjust for send format´byte order
            if (str.Length != 4)
            {
                throw new Exception("Wrong length");
            }
            return Int16.Parse(str.Substring(2, 2) + str.Substring(0, 2), System.Globalization.NumberStyles.HexNumber);
        }

        public static double hexToVolt(string str, voltages volt)
        {
            return MyECUData.valueToVolt(hexToInt(str), volt);
        }

        public static double valueToVolt(int val, voltages volt)
        {
            if (val < 0)
            {
                val = 0 - val;
            }
            switch(volt)
            { 
                case voltages.Volt5:
                   return (5.0 / 1024) * val;
                case voltages.Volt12:
                   return (5.0 / 1024) * val * MyECUSource.vccmult;
                default:
                   throw new Exception("Unhandled voltage");
            }
        }

        public static double hexToTemp(string str)
        {
            return valueToTemp(hexToUInt(str));
        }

        public static double valueToTemp(uint vbridge)
        {
            double vbridgeVolt = valueToVolt((int) vbridge, voltages.Volt5);
            double rtherm = vbridgeVolt / ((5.0 - vbridgeVolt) / MyECUSource.rfix);
            double divider =      MyECUSource.SHCoeffs[0] 
                                + MyECUSource.SHCoeffs[1] * Math.Log(rtherm) 
                                + MyECUSource.SHCoeffs[2]*Math.Pow(Math.Log(rtherm),2)
                                + MyECUSource.SHCoeffs[3]*Math.Pow(Math.Log(rtherm),3);
            double temp = (1 / divider) - 273.15;
            return temp;
        }

        public MyECUData(String line)
        {
            // When we arrive here the data is already trimmed from leading and trailing whitespace
            String pat = @"^=?([013d])([0-9A-F]{2})([0-9A-F]+)([0-9A-F]{2})$";
            Regex r = new Regex(pat);
            Match m = r.Match(line);
            if (m.Success)
            {
                string typeString = m.Groups[1].ToString();
                string OctetsNumString = m.Groups[2].ToString();
                string dataString = m.Groups[3].ToString();
                string checkSumString = m.Groups[4].ToString();

                int numOctets = Int16.Parse(OctetsNumString, System.Globalization.NumberStyles.HexNumber);
                int checkSum = Int16.Parse(checkSumString, System.Globalization.NumberStyles.HexNumber);
                int typeNr = Int16.Parse(typeString, System.Globalization.NumberStyles.HexNumber);
                int checkSumCalc = 0;

                if (2 * numOctets == dataString.Length)
                {
                    for (int i = 0; i < numOctets; i++)
                    {
                        int octet = Int16.Parse(dataString.Substring(i * 2, 2), System.Globalization.NumberStyles.HexNumber);
                        checkSumCalc = (checkSumCalc + octet) & 0xFF;
                    }
                    if (checkSumCalc == checkSum)
                    {
                        switch (typeNr)
                        {
                            case 0:
                                ai = new MyECUDataAnalogInputs(line, dataString);
                                type = types.AnalogInputs;
                                break;
                            case 1:
                                ti = new MyECUDataTimingInfo(line, dataString);
                                type = types.TimingInfo;
                                break;
                            case 3:
                                ci = new MyECUDataCompensationInfo(line, dataString);
                                type = types.CompensationInfo;
                                break;
                            case 0x0d:
                                di = new MyECUDataDiagnosticInfo(line, dataString);
                                type = types.DiagnosticInfo;
                                break;
                            default:
                                break;
                        }
                    }
                }
            }
            else
            {
                // Catch version strings like "~My16M2 v4.2 E609"
                pat = @"^~(My.*)$";
                r = new Regex(pat);
                m = r.Match(line);
                if (m.Success)
                {
                    vi = new MyECUDataVersionInfo(line, m.Groups[1].ToString());
                    type = types.VersionInfo;
                }
            }
        }
    }
}



/*
Map (7.8,4.1) RPM=3242 Inj=5977us BaroAdj=106.0% o2Adj=102.2% o2[3] o2Target=2.39V o2=2.09V OilT=80.0C AirT=11.0C AirP=1020mb Thro=372 Bat=13.92V
=3 10 FC03  DB43   17B   4 EA01 EA01 0004 3717 17B    4 DB
      1020 17371  2329 4 0  490  490 1024 5943 2329 4 0
      mBar 106.0 102.2     2.39 2.39
 
=112580113290010D9023010301000001E2845D25D
=3 12 EE03 4241 009 4 0000 0000 0004 E60F 0094 000 0 95
 -  -  ?     ?    ? ?  ?     ?    ?   ?     ?   ?  ? -
=00E7E02E200E4029F03D600F6011000C7
Map (8.8,2.3) Thro=214 RPM=3429 o2Adj=100.0%,100.0% o2[1] o2Target=0.00V o2=3.12V,0.08V OilT=76.0C AirT=18.0C AirP=1006mb Bat=13.97V Inj=4070us BaroAdj=102.0%
 
=112030213920010030359175917000023470DC5DC
=310FC03DB4317B4EA01EA010004371717B4DB
=00EAD01CA002303A6037401F4014601F8
Map (7.8,4.1) RPM=3242 Inj=5977us BaroAdj=106.0% o2Adj=102.2% o2[3] o2Target=2.39V o2=2.09V OilT=80.0C AirT=11.0C AirP=1020mb Thro=372 Bat=13.92V


=1127F0014D402102E038F2F8F2F000023B73F68A7
=3 10 FC03 4C44 0094000000000004992F009483
=00E7A011F013403A903DB02F4019000E0
Map (7.4,11.2) RPM=3071 Inj=12175us BaroAdj=106.7% o2Adj=100.0% o2[1] o2Target=0.00V o2=1.85V OilT=67.0C AirT=9.0C AirP=1020mb Thro=731 Bat=13.92V

(RPM,THR)

Below is a list is the commands and response packets that My16M uses. All command and responses are alphanumeric and are 
terminated with a carriage return - '\r' which has a binary value of 13. Multi-byte values are transferred little endian 
(low byte first).

Memory structure transfer
=========================
Many of the packets are used to transfer internal ECU structures. These structures are transferred using a common 
subroutine. A memory structure is transferred in the flowing format.

<prefix>NND0D1...DnXX

where
	NN is the number of bytes in the structure in hex
	D0..Dn are the memory bytes of the structure
	XX is the checksum


For those interested the following routine is used to do transfer the structure

		void TxMem_( uint8_t * Buf, int Cnt, char * Str ) {
				int i;
				uint8_t Chk = 0;
				TxStr_P( Str );
				Txu8( Cnt );
				for ( i = 0; i < Cnt; i++ ) {
						int i;
						uint8_t byt = Buf[i];
						Txu8( byt );
						Chk += byt;
				}
				Txu8( Chk );
				TxStr_P(CR);		
		}

The Str parameter is used to prefix the memory structure for identification.

Continuous Updates
==================
As soon as the ECU is powered up it begins dumping 3 structures continuously. The rate of transfer is RPM dependant but 
for all intents and purposes it can be consider realtime ( 5 - 10 times per second ).
*/