﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;

namespace MyECULib
{
    public class MyECUDataCompensationInfo
    {
        public string source;
        public uint AirPressure;
        public double BaroAdjust; // The correction factor calculated from air pressure and temperature 1024 0x0400BE is 100% - no change from map
        public double O2Adjust0; // The correction factor being applied to the map ( after baro correction ) 12 Bit
        public uint O2Index0; // 3 Bit 
        public uint ClosedLoopActive0; // 1 Bit
        public double O2Target0;
        public double O2Target1;
        public double CylinderOffSet; // 1024 = 1 = 100%
        public uint InjDuration0; // Pulse duration in us
        public uint InjDuration1; // Will be calculated from the two above
        public double O2Adjust1; // The correction factor being applied to the map ( after baro correction ) 12 bits
        public uint O2Index1;  // 3 bit
        public uint ClosedLoopActive1; // 1 bit
        public uint InjDeadTime; // Injector dead time
        public bool InjDeadTimePresent;

        private static Regex reg = new Regex(@"([0-9A-F]{4})([0-9A-F]{4})([0-9A-F]{4})([0-9A-F]{4})([0-9A-F]{4})([0-9A-F]{4})([0-9A-F]{4})([0-9A-F]{4})([0-9A-F]{4})?$");

        private static void unpackOverloadedInfo(uint data, ref double adjust, ref uint index, ref uint active)
        {
            adjust = (data & 2047);
            adjust = (adjust/1024);
            index = (data >> 12) & 7;
            active = (data >> 15);
        }

        override public string ToString()
        {
            string res = "CompensationInfo " + source + "\n" +
            "AirP: " + AirPressure + " mBar\n" +
            "BaroAdj: " + System.Math.Round(BaroAdjust * 100, 1) + " %\n" +
            "O2Adj0: " + System.Math.Round(O2Adjust0 * 100, 1) + " %\n" +
            "O2Adj1: " + System.Math.Round(O2Adjust1 * 100, 1) + " %\n" +
            "CL0: " + ClosedLoopActive0 + ", O2[" + O2Index0 + "], " + System.Math.Round(O2Target0, 2) + 
            " V => Lambda: " + System.Math.Round(MyECUData.voltage2Lambda(O2Target0), 2) + 
            " => AFR: " + System.Math.Round(MyECUData.lambda2AFR(MyECUData.voltage2Lambda(O2Target0)), 2) + "\n" +
            "CL1: " + ClosedLoopActive1 + ", O2[" + O2Index1 + "], " + System.Math.Round(O2Target1, 2) +
            " V => Lambda: " + System.Math.Round(MyECUData.voltage2Lambda(O2Target1), 2) +
            " => AFR: " + System.Math.Round(MyECUData.lambda2AFR(MyECUData.voltage2Lambda(O2Target1)), 2) + "\n" +
            "MapInjDur0: " + System.Math.Round((double)InjDuration0 / BaroAdjust, 0) + " us\n" +
            "MapInjDur1: " + System.Math.Round((double)InjDuration1 / BaroAdjust, 0) + " us\n" +
            "CylinderOffset: " + (System.Math.Round(CylinderOffSet * 100, 1) - 100) + "%\n" +
            "InjDur0: " + InjDuration0 + " us";
            if (O2Adjust0 != 1.0)
            {
                res += "  =>  " + System.Math.Round((double) InjDuration0 * O2Adjust0,0) + " us incl O2Adj";
            }
            res += "\n";
            if (InjDeadTimePresent)
            {
                res += "InjDeadTime: " + InjDeadTime + " us\n";
            }
            res += "InjDur1: " + InjDuration1 + " us";
            if (O2Adjust1 != 1.0)
            {
                res += "  =>  " + System.Math.Round((double)InjDuration1 * O2Adjust1, 0) + " us incl O2Adj";
            }
            res += "\n";
            if (InjDeadTimePresent)
            {
                res += "InjDeadTime: " + InjDeadTime + " us\n";
            }
            res += "\n";
            return res;
        }

        public MyECUDataCompensationInfo(String line, String data)
        {
            source = line;
            Match m = reg.Match(data);
            if (m.Success)
            {
                uint shared1;
                uint shared2;

                AirPressure = MyECUData.hexToUInt(m.Groups[1].ToString());
                string bas = m.Groups[2].ToString();
                double ba = (double) (MyECUData.hexToUInt(bas));
                BaroAdjust = (ba / (16 * 1024));  // Documentation sucks - we need a /16
                shared1 = MyECUData.hexToUInt(m.Groups[3].ToString());
                unpackOverloadedInfo(shared1, ref O2Adjust0, ref O2Index0, ref ClosedLoopActive0);

                O2Target0 = MyECUData.valueToVolt(MyECUData.hexToInt(m.Groups[4].ToString()), MyECUData.voltages.Volt5);
                O2Target1 = MyECUData.valueToVolt(MyECUData.hexToInt(m.Groups[5].ToString()), MyECUData.voltages.Volt5);
                CylinderOffSet = (double)(MyECUData.hexToUInt(m.Groups[6].ToString()) / 1024);
                InjDuration0 = MyECUData.hexToUInt(m.Groups[7].ToString());
                InjDuration1 = (uint)((double)InjDuration0 * CylinderOffSet);
                shared2 = MyECUData.hexToUInt(m.Groups[8].ToString());
                unpackOverloadedInfo(shared2, ref O2Adjust1, ref O2Index1, ref ClosedLoopActive1);

                string tmp = m.Groups[9].ToString();
                if (tmp.Length > 0)
                {
                    InjDeadTime = MyECUData.hexToUInt(m.Groups[9].ToString());
                    InjDeadTimePresent = true;
                }
                else
                {
                    InjDeadTimePresent = false;
                    InjDeadTime = 0;
                }
               
               
            }
            else
            {
                //throw new Exception("Bad data");
            }
        }

    }
}
/*
Compensation Info
=================
This structure can be used to determine various corrections being applied to the map
This structure is prefixed with an "=3".

		typedef struct {
				uint16_t AirPressure;
				uint16_t BaroAdjust; // The correction factor calculated from air pressure and temperature 1024 0x0400BE is 100% - no change from map
				uint16_t O2Adjust0:12; // The correction factor being applied to the map ( after baro correction )
				uint16_t O2Index0:3;
				uint16_t ClosedLoopActive0:1;
				int16_t O2Target0;
				int16_t O2Target1;
				uint16_t CylinderOffSet;
				uint16_t InjDuration; // Pulse duration in us
				uint16_t O2Adjust1:12; // The correction factor being applied to the map ( after baro correction )
				uint16_t O2Index1:3;
				uint16_t ClosedLoopActive1:1;
				uint16_t InjectorVAdjust;
		} CompInfo_t;

Map (7.8,4.1) RPM=3242 Inj=5977us BaroAdj=106.0% o2Adj=102.2% o2[3] o2Target=2.39V o2=2.09V OilT=80.0C AirT=11.0C AirP=1020mb Thro=372 Bat=13.92V
=3 10 FC03  DB43   17B   4 EA01 EA01 0004 3717 17B    4 DB
      1020 17371  2329 4 0  490  490 1024 5943 2329 4 0
      mBar 106.0 102.2     2.39 2.39

                                        ~1093 = 0x4504LE
Map (7.4,11.2) RPM=3071 Inj=12175us BaroAdj=106.7% o2Adj=100.0% o2[1] o2Target=0.00V o2=1.85V OilT=67.0C AirT=9.0C AirP=1020mb Thro=731 Bat=13.92V
=3 10 FC03  4C44  009   4 0000 0000  0004  992F  009   4 83	hex LE
      1020 17484 2304 4 0    0    0  1024 12185	2304 4 0		dec
      mBar 106.7          0.00 0.00

=3 10 03FC  444C  940   0 0000 0000  0400  2F99  940   0 83	hex BE

new:
Map (0.0,12.8) Thro=459 RPM=200 o2Adj=100.0%,100.0% o2[0] o2Target=0.00V o2=0.02V,2.24V OilT=24.0C AirT=22.0C AirP=990mb Bat=12.19V Inj=0us BaroAdj=110.0%
=3 12 DE03  6846 000   4 0000 0000 0004 0000 000   4 0000 9B
       990 18024   0 4 0    0    0 1024    0   0 4 0    0
*/